#  Copyright 2021 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import warnings
from datetime import datetime
from enum import Enum
from typing import Optional

import jwt
import requests
from dataclassy import dataclass
from urllib3.exceptions import InsecureRequestWarning

from zepben.auth.util import construct_url


__all__ = ["ZepbenAuthenticator", "AuthMethod", "AuthException", "create_authenticator"]

_AUTH_HEADER_KEY = 'authorization'


class AuthException(Exception):
    pass


class AuthMethod(Enum):
    """
    An enum class that represents the different authentication methods that could be returned from the server's
    ewb/config/auth endpoint.
    """
    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.upper():
                return member

    NONE = "NONE"
    SELF = "self"
    AUTH0 = "AUTH0"


@dataclass
class ZepbenAuthenticator(object):
    audience: str
    """ Audience to use when requesting tokens """

    issuer_domain: str
    """ The domain of the token issuer. """

    auth_method: AuthMethod
    """ The authentication method used by the server """

    verify_certificate: bool = False
    """ Whether to verify the SSL certificate when making requests """

    issuer_protocol: str = "https"
    """ Protocol of the token issuer. You should not change this unless you are absolutely sure of what you are doing. Setting it to
        anything other than https is a major security risk as tokens will be sent in the clear. """

    token_path: str = "/oauth/token"
    """ Path for requesting token from `issuer_domain`. """

    algorithm: str = "RS256"
    """ Algorithm used for decoding tokens. Note tokens are only decoded for checking expiry time."""

    token_request_data = {}
    """ Data to pass in token requests. """

    refresh_request_data = {}
    """ Data to pass in refresh token requests. """

    _access_token = None
    _refresh_token = None
    _token_expiry = datetime.min
    _token_type = None

    def __init__(self):
        self.token_request_data["audience"] = self.audience
        self.refresh_request_data["audience"] = self.audience

    def fetch_token(self) -> str:
        """
        Returns a JWT access token and its type in the form of '<type> <3 part JWT>', retrieved from the configured OAuth2 token provider.
        Throws AuthException if an access token request fails.
        """
        if datetime.utcnow() > self._token_expiry:
            # Stored token has expired, try to refresh
            self._access_token = None
            if self._refresh_token:
                self._fetch_token_auth0(use_refresh=True)

            if self._access_token is None:
                # If using the refresh token did not work for any reason, self._access_token will still be None.
                # and thus we must try get a fresh access token using credentials instead.
                self._fetch_token_auth0()

            # Just to give a friendly error if a token retrieval failed for a case we haven't handled.
            if not self._token_type or not self._access_token:
                raise AuthException(
                    f"Token couldn't be retrieved from {construct_url(self.issuer_protocol, self.issuer_domain, self.token_path)} using configuration "
                    f"{self.auth_method}, audience: {self.audience}, token issuer: {self.issuer_domain}")

        return f"{self._token_type} {self._access_token}"

    def _fetch_token_auth0(self, use_refresh: bool = False):
        response = requests.post(
            construct_url(self.issuer_protocol, self.issuer_domain, self.token_path),
            headers={"content-type": "application/x-www-form-urlencoded"},
            data=self.refresh_request_data if use_refresh else self.token_request_data,
            verify=self.verify_certificate
        )

        if not response.ok:
            raise AuthException(f'Token fetch failed, Error was: {response.status_code} - {response.reason} {response.text}')

        try:
            data = response.json()
        except ValueError as e:
            raise AuthException(f'Response did not contain expected JSON - response was: {response.text}', e)

        if "error" in data or "access_token" not in data:
            raise AuthException(f'{data.get("error", "Access Token absent in token response")} - {data.get("error_description", f"Response was: {data}")}')

        self._token_type = data["token_type"]
        self._access_token = data["access_token"]
        self._token_expiry = datetime.fromtimestamp(jwt.decode(self._access_token, algorithms=["RS256"], options={"verify_signature": False})['exp'])

        if use_refresh:
            self._refresh_token = data["refresh_token"]


def create_authenticator(conf_address: str, verify_certificate: bool = True, auth_type_field: str = 'authType',
                         audience_field: str = 'audience', issuer_domain_field: str = 'issuer') -> Optional[ZepbenAuthenticator]:
    """
    Helper method to fetch auth related configuration from `conf_address` and create a :class:`ZepbenAuthenticotar`

    :param conf_address: Location to retrieve authentication configuration from. Must be a HTTP address that returns a JSON response.
    :param verify_certificate: Whether to verify the certificate when making HTTPS requests. Note you should only use a trusted server
        and never set this to False in a production environment.
    :param auth_type_field: The field name to look up in the JSON response from the conf_address for `authenticator.auth_method`.
    :param audience_field: The field name to look up in the JSON response from the conf_address for `authenticator.auth_method`.
    :param issuer_domain_field: The field name to look up in the JSON response from the conf_address for `authenticator.auth_method`.

    :returns: A :class:`ZepbenAuthenticator` if the server reported authentication was configured, otherwise None.
    """
    with warnings.catch_warnings():
        if not verify_certificate:
            warnings.filterwarnings("ignore", category=InsecureRequestWarning)
        response = requests.get(conf_address, verify=verify_certificate)
        if response.ok:
            try:
                auth_config_json = response.json()
                auth_method = AuthMethod(auth_config_json[auth_type_field])
                if auth_method is not AuthMethod.NONE:
                    return ZepbenAuthenticator(
                        auth_config_json[audience_field],
                        auth_config_json[issuer_domain_field],
                        auth_method,
                        verify_certificate
                    )
            except ValueError as e:
                raise ValueError(f"Expected JSON response from {conf_address}, but got: {response.text}.", e)
        else:
            raise ValueError(f"{conf_address} responded with error: {response.status_code} - {response.reason} {response.text}")
    return None
